# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import logging
import yaml
from .requirements import RequirementsBundle
from .providers.github import Provider as GithubProvider
from .errors import NoPermissionError, BranchExistsError, ConfigError
from .config import Config

logger = logging.getLogger(__name__)


class Bot(object):
    def __init__(self, repo, user_token, bot_token=None,
                 provider=GithubProvider, bundle=RequirementsBundle, config=Config,
                 integration=False, provider_url=None, ignore_ssl=False):
        self.req_bundle = bundle()
        self.provider = provider(self.req_bundle, integration, provider_url, ignore_ssl)
        self.user_token = user_token
        self.bot_token = bot_token
        self.fetched_files = []
        self.repo_name = repo

        self._user = None
        self._user_repo = None
        self._bot = None
        self._bot_repo = None
        self.config = config()
        self.write_config = {}

        self._fetched_prs = False

        self.integration = integration

    @property
    def user_repo(self):
        if self._user_repo is None:
            self._user_repo = self.provider.get_repo(token=self.user_token, name=self.repo_name)
        return self._user_repo

    @property
    def user(self):
        if self._user is None:
            self._user = self.provider.get_user(token=self.user_token)
        return self._user

    @property
    def bot(self):
        if self._bot is None:
            self._bot = self.provider.get_user(token=self.bot_token)
        return self._bot

    @property
    def bot_repo(self):
        if self._bot_repo is None:
            self._bot_repo = self.provider.get_repo(token=self.bot_token, name=self.repo_name)
        return self._bot_repo

    @property
    def pull_requests(self):
        if not self._fetched_prs:
            self.req_bundle.pull_requests = [
                pr for pr in self.provider.iter_issues(
                    repo=self.user_repo,
                    creator=self.bot if self.bot_token else self.user
                )
                if pr.is_valid
            ]
            self._fetched_prs = True
        return self.req_bundle.pull_requests

    def get_repo_config(self, repo, branch=None, create_error_issue=True):
        branch = self.config.branch if branch is None else branch
        content, _ = self.provider.get_file(repo, ".pyup.yml", branch)
        if content is not None:
            try:
                return yaml.safe_load(content)
            except yaml.YAMLError as e:
                err = ConfigError(content=content, error=e.__str__())
                if create_error_issue:
                    issue_title = "Invalid .pyup.yml detected"
                    # check that there's not an open issue already
                    if issue_title not in [pr.title for pr in self.pull_requests if pr.is_open]:
                        self.create_issue(
                            title=issue_title,
                            body="The bot encountered an error in your `.pyup.yml` config file:\n\n"
                                 "```{error}\n```\n\n"
                                 "You can validate it with this "
                                 "[online YAML parser](http://yaml-online-parser.appspot.com/) or "
                                 "by taking a look at the "
                                 "[Documentation](https://pyup.io/docs/bot/config/).".format(
                                    error=err.error)
                        )
                raise err
        return None

    def configure(self, create_error_issue=True, **kwargs):
        if kwargs.get("write_config", False):
            self.write_config = kwargs.get("write_config")
        # if the branch is not set, get the default branch
        if kwargs.get("branch", False) in [None, False]:
            self.config.branch = self.provider.get_default_branch(repo=self.user_repo)
        # set the config for this update run
        self.config.update_config(kwargs)
        repo_config = self.get_repo_config(
            repo=self.user_repo,
            create_error_issue=create_error_issue
        )
        if repo_config:
            self.config.update_config(repo_config)
        if self.write_config:
            self.config.update_config(self.write_config)
        logger.info("Runtime config is: {}".format(self.config))

    def update(self, **kwargs):
        """
        Main entrypoint to kick off an update run.
        :param kwargs:
        :return: RequirementsBundle
        """
        self.configure(**kwargs)
        self.get_all_requirements()
        self.apply_updates(
            initial=kwargs.get("initial", False),
            scheduled=kwargs.get("scheduled", False)
        )

        return self.req_bundle

    def can_pull(self, initial, scheduled):
        """
        Determines if pull requests should be created
        :return: bool
        """
        if not initial and self.config.is_valid_schedule():
            # if the config has a valid schedule, return True if this is a scheduled run
            return scheduled
        return True

    def apply_updates(self, initial, scheduled):

        InitialUpdateClass = self.req_bundle.get_initial_update_class()

        if initial:
            # get the list of pending updates
            try:
                _, _, _, updates = list(
                    self.iter_updates(initial, scheduled)
                )[0]
            except IndexError:
                # need to catch the index error here in case the intial update is completely
                # empty
                updates = False
            # if this is the initial run and the update list is empty, the repo is already
            # up to date. In this case, we create an issue letting the user know that the bot is
            # now set up for this repo and return early.
            if not updates:
                title = InitialUpdateClass.get_title()
                if self.config.pr_prefix:
                    title = "{prefix} {title}".format(prefix=self.config.pr_prefix, title=title)
                self.create_issue(
                    title=title,
                    body=InitialUpdateClass.get_empty_update_body()
                )
                if self.write_config:
                    self.pull_config(self.write_config)
                return

        # check if we have an initial PR open. If this is the case, we attach the initial PR
        # to all updates and are done. The `Initial Update` has to be merged (or at least closed)
        # before we continue to do anything here.
        initial_pr = next(
            (pr for pr in self.pull_requests if
             pr.canonical_title(self.config.pr_prefix) ==
             InitialUpdateClass.get_title() and pr.is_open),
            False
        )

        if initial and self.write_config:
            self.pull_config(self.write_config)

        # todo: This block needs to be refactored
        for title, body, update_branch, updates in self.iter_updates(initial, scheduled):
            # some scheduled updates don't have commits in them. This happens if a package is
            # outdated, but the config file is blocking the update (insecure, no updates).
            # check if this is the case here.
            if not updates:
                continue

            if self.config.pr_prefix:
                title = "{prefix} {title}".format(prefix=self.config.pr_prefix, title=title)
            if initial_pr:
                pull_request = initial_pr
            elif self.can_pull(initial, scheduled) and \
                    title not in [pr.title for pr in self.pull_requests]:
                update_branch = self.config.branch_prefix + update_branch
                pull_request = self.commit_and_pull(
                    initial=initial,
                    new_branch=update_branch,
                    title=title,
                    body=body,
                    updates=updates,
                )
            else:
                pull_request = next((pr for pr in self.pull_requests if pr.title == title), None)

            logger.info("Have updates {} and pr {}".format(updates, pull_request))
            for update in updates:
                update.requirement.pull_request = pull_request
                if self.config.close_prs and pull_request and not initial:
                    self.close_stale_prs(
                        update=update,
                        pull_request=pull_request,
                        scheduled=scheduled
                    )
                    # if this is a scheduled update, break since it's bundled
                    if pull_request.is_scheduled:
                        break

    def close_stale_prs(self, update, pull_request, scheduled):
        """
        Closes stale pull requests for the given update, links to the new pull request and deletes
        the stale branch.
        A stale PR is a PR that:
         - Is not merged
         - Is not closed
         - Has no commits (except the bot commit)
        :param update:
        :param pull_request:
        """
        closed = []
        if self.bot_token and not pull_request.is_initial:
            for pr in self.pull_requests:
                close_pr = False
                same_title = \
                    pr.canonical_title(self.config.pr_prefix) == \
                    pull_request.canonical_title(self.config.pr_prefix)

                if scheduled and pull_request.is_scheduled:
                    # check that the PR is open and the title does not match
                    if pr.is_open and not same_title:
                        # we want to close the previous scheduled PR if it is not merged yet
                        # and we want to close all previous updates if the user choose to
                        # switch to a scheduled update
                        if pr.is_scheduled or pr.is_update:
                            close_pr = True
                elif pull_request.is_update:
                    # check that, the pr is an update, is open, the titles are not equal and that
                    # the requirement matches
                    if pr.is_update and \
                        pr.is_open and \
                            not same_title and \
                            pr.get_requirement(self.config.pr_prefix) == update.requirement.key:
                        # there's a possible race condition where multiple updates with more than
                        # one target version conflict with each other (closing each others PRs).
                        # Check that's not the case here
                        if not self.has_conflicting_update(update):
                            close_pr = True

                if close_pr and self.is_bot_the_only_committer(pr=pr):
                    logger.info("Closing stale PR {} for {}".format(pr.title, pull_request.title))
                    self.provider.close_pull_request(
                        bot_repo=self.bot_repo,
                        user_repo=self.user_repo,
                        pull_request=pr,
                        comment="Closing this in favor of #{}".format(
                            pull_request.number),
                        prefix=self.config.branch_prefix
                    )
                    pr.state = "closed"
                    closed.append(pr)
        for closed_pr in closed:
            self.pull_requests.remove(closed_pr)

    def is_bot_the_only_committer(self, pr):
        """
        Checks if the bot is the only committer for the given pull request.
        :param update: Update to check
        :return: bool - True if conflict found
        """
        committer = self.provider.get_pull_request_committer(
            self.user_repo,
            pr)
        # flatten the list and remove duplicates
        committer_set = set([c.login for c in committer])

        # it's impossible to get the bots login if this is an integration, just check that
        # there's only one commit in the commit history.
        if self.integration or getattr(self.provider, 'name', '') == 'gitlab':
            return len(committer_set) == 1

        # check that there's exactly one committer in this PRs commit history and
        # that the committer is the bot
        return len(committer_set) == 1 and self.provider.is_same_user(self.bot, committer[0])

    def has_conflicting_update(self, update):
        """
        Checks if there are conflicting updates. Conflicting updates are updates that have the
        same requirement but different target versions to update to.
        :param update: Update to check
        :return: bool - True if conflict found
        """
        # we explicitly want a flat list of updates here, that's why we call iter_updates
        # with both `initial` and `scheduled` == False
        for _, _, _, updates in self.iter_updates(initial=False, scheduled=False):
            for _update in updates:
                if (
                        update.requirement.key == _update.requirement.key and
                        (
                                update.commit_message != _update.commit_message or
                                update.requirement.latest_version_within_specs !=
                                _update.requirement.latest_version_within_specs
                        )
                ):
                    logger.info("{} conflicting with {}/{}".format(
                        update.requirement.key,
                        update.requirement.latest_version_within_specs,
                        _update.requirement.latest_version_within_specs)
                    )
                    return True
        return False

    def create_branch(self, new_branch, delete_empty=False):
        """
        Creates a new branch.
        :param new_branch: string name of the new branch
        :param delete_empty: bool -- delete the branch if it is empty
        :return: bool -- True if successfull
        """
        logger.info("Preparing to create branch {} from {}".format(new_branch, self.config.branch))
        try:
            # create new branch
            self.provider.create_branch(
                base_branch=self.config.branch,
                new_branch=new_branch,
                repo=self.user_repo
            )
            logger.info("Created branch {} from {}".format(new_branch, self.config.branch))
            return True
        except BranchExistsError:
            logger.info("Branch {} exists.".format(new_branch))
            # if the branch exists, is empty and delete_empty is set, delete it and call
            # this function again
            if delete_empty:
                if self.provider.is_empty_branch(self.user_repo, self.config.branch, new_branch,
                                                 self.config.branch_prefix):
                    self.provider.delete_branch(self.user_repo, new_branch,
                                                self.config.branch_prefix)
                    logger.info("Branch {} was empty and has been deleted".format(new_branch))
                    return self.create_branch(new_branch, delete_empty=False)
                logger.info("Branch {} is not empty".format(new_branch))
        return False

    def pull_config(self, new_config):  # pragma: no cover
        """

        :param new_config:
        :return:
        """
        logger.info("Creating new config file with {}".format(new_config))
        branch = 'pyup-config'
        if self.create_branch(branch, delete_empty=True):
            content = self.config.generate_config_file(new_config)
            _, content_file = self.provider.get_file(self.user_repo, '.pyup.yml', branch)
            if content_file:
                # a config file exists, update and commit it
                logger.info(
                    "Config file exists, updating config for sha {}".format(content_file.sha))
                self.provider.create_commit(
                    repo=self.user_repo,
                    path=".pyup.yml",
                    branch=branch,
                    content=content,
                    commit_message="update pyup.io config file",
                    committer=self.bot if self.bot_token else self.user,
                    sha=content_file.sha
                )
            logger.info("No config file found, writing new config file")
            # there's no config file present, write a new config file and commit it
            self.provider.create_and_commit_file(
                repo=self.user_repo,
                path=".pyup.yml",
                branch=branch,
                content=content,
                commit_message="create pyup.io config file",
                committer=self.bot if self.bot_token else self.user,
            )

            title = 'Config file for pyup.io'
            if self.config.pr_prefix:
                title = "{prefix} {title}".format(prefix=self.config.pr_prefix, title=title)
            body = 'Hi there and thanks for using pyup.io!\n' \
                   '\n' \
                   "Since you are using a non-default config I've created one for you.\n\n" \
                   "There are a lot of things you can configure on top of " \
                   "that, so make sure to check out the " \
                   "[docs](https://pyup.io/docs/configuration/) to see what I can do for you."

            pr = self.create_pull_request(
                title=title,
                body=body,
                new_branch=branch,
            )
            return pr

    def commit_and_pull(self, initial, new_branch, title, body, updates):
        logger.info("Preparing commit {}".format(title))
        if self.create_branch(new_branch, delete_empty=False):
            updated_files = {}
            for update in self.iter_changes(initial, updates):
                if update.requirement_file.path in updated_files:
                    sha = updated_files[update.requirement_file.path]["sha"]
                    content = updated_files[update.requirement_file.path]["content"]
                else:
                    sha = update.requirement_file.sha
                    content = update.requirement_file.content
                old_content = content
                content = update.requirement.update_content(content, self.config.update_hashes)
                if content != old_content:
                    new_sha = self.provider.create_commit(
                        repo=self.user_repo,
                        path=update.requirement_file.path,
                        branch=new_branch,
                        content=content,
                        commit_message=update.commit_message,
                        sha=sha,
                        committer=self.bot if self.bot_token else self.user,
                    )
                    updated_files[update.requirement_file.path] = {"sha": new_sha,
                                                                   "content": content}
                else:
                    if hasattr(self.user_repo, 'path_with_namespace'):
                        repo_name = self.user_repo.path_with_namespace
                    elif hasattr(self.user_repo, 'full_name'):
                        repo_name = self.user_repo.full_name
                    else:
                        repo_name = str(self.user_repo)
                    logger.error("Empty commit at {repo}, unable to update {title}.".format(
                        repo=repo_name, title=title)
                    )

            if updated_files:
                pr = self.create_pull_request(
                    title=title,
                    body=body,
                    new_branch=new_branch,
                )
                self.pull_requests.append(pr)
                return pr
        return None

    def create_issue(self, title, body):
        return self.provider.create_issue(
            repo=self.bot_repo if self.bot_token else self.user_repo,
            title=title,
            body=body,
        )

    def create_pull_request(self, title, body, new_branch):

        # if we have a bot user that creates the PR, we might run into problems on private
        # repos because the bot has to be a collaborator. We try to submit the PR before checking
        # the permissions because that saves us API calls in most cases
        kwargs = dict(title=title,body=body, new_branch=new_branch,
                      base_branch=self.config.branch, pr_label=self.config.label_prs, assignees=self.config.assignees,
                      config=self.config)
        if self.bot_token:
            try:
                return self.provider.create_pull_request(repo=self.bot_repo, **kwargs)
            except NoPermissionError:
                self.provider.get_pull_request_permissions(self.bot, self.user_repo)

        return self.provider.create_pull_request(
            repo=self.bot_repo if self.bot_token else self.user_repo,
            **kwargs
        )

    def iter_git_tree(self, sha=None):
        branch = sha if sha is not None else self.config.branch
        return self.provider.iter_git_tree(branch=branch, repo=self.user_repo)

    def iter_updates(self, initial, scheduled):
        return self.req_bundle.get_updates(
            initial=initial,
            scheduled=scheduled,
            config=self.config
        )

    def iter_changes(self, initial, updates):
        return iter(updates)

    # if this function gets updated, the gist at https://gist.github.com/jayfk/45862b05836701b49b01
    # needs to be updated too
    def get_all_requirements(self, sha=None):
        if self.config.search:
            logger.info("Searching requirement files")
            for file_type, path in self.iter_git_tree(sha=sha):
                if file_type == "blob":
                    if "requirements" in path:
                        if path.endswith("txt") or path.endswith("pip"):
                            self.add_requirement_file(path, sha)
                    if "setup.cfg" in path:
                        self.add_requirement_file(path, sha)
        for req_file in self.config.requirements:
            self.add_requirement_file(req_file.path, sha=sha)
        self.req_bundle.resolve_pipfiles()

    # if this function gets updated, the gist at https://gist.github.com/jayfk/c6509bbaf4429052ca3f
    # needs to be updated too
    def add_requirement_file(self, path, sha=None):
        logger.info("Adding requirement file at {}".format(path))
        branch = sha if sha is not None else self.config.branch
        if not self.req_bundle.has_file_in_path(path):
            req_file = self.provider.get_requirement_file(
                path=path, repo=self.user_repo, branch=branch)
            if req_file is not None:
                self.req_bundle.append(req_file)
                for other_file in req_file.other_files:
                    self.add_requirement_file(other_file, sha=sha)


class DryBot(Bot):
    def commit_and_pull(self, initial, new_branch, title, body, updates):  # pragma: no cover
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import os
from collections import namedtuple
from datetime import datetime
from jinja2 import Environment, FileSystemLoader
from .errors import UnsupportedScheduleError

from pyup import settings

TEMPLATES_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "templates"
)


class Update(dict):

    @classmethod
    def create_update_key(cls, requirement):
        key = requirement.key
        if not requirement.is_pinned:
            key += '-pin'
        else:
            key += "-" + requirement.latest_version_within_specs
        return key

    def __init__(self, requirement_files, config):
        super(dict, self).__init__()
        self.config = config
        for requirement_file in requirement_files:
            for requirement in requirement_file.requirements:
                if requirement.needs_update:
                    self.add(requirement, requirement_file)

    def add(self, requirement, requirement_file):
        key = self.create_update_key(requirement)
        update = RequirementUpdate(
            requirement=requirement,
            requirement_file=requirement_file,
            commit_message=self.get_commit_message(requirement)
        )
        if key in self:
            self[key].append(update)
        else:
            self[key] = [update]

    @classmethod
    def get_commit_message(cls, requirement):
        if requirement.is_pinned:
            return "Update {} from {} to {}".format(
                requirement.key, requirement.version,
                requirement.latest_version_within_specs
            )
        return "Pin {} to latest version {}".format(
            requirement.key,
            requirement.latest_version_within_specs
        )

    def should_update(self, requirement, requirement_file):
        """
        Determines if a requirement can be updated
        :param requirement: Requirement
        :param requirement_file: RequirementFile
        :return: bool
        """
        path = requirement_file.path
        if self.config.can_update_all(path) or \
                (self.config.can_update_insecure(path) and requirement.is_insecure):
            # handle unpinned requirements only if pin is set
            if not requirement.is_pinned:
                return self.config.can_pin(path)
            return True
        return False

    def get_requirement_update_class(self):
        return RequirementUpdate


class BundledUpdate(Update):

    def get_updates(self):
        if self:
            yield (
                self.get_title(),
                self.get_body([update for updates in self.values() for update in updates
                               if
                               self.should_update(update.requirement, update.requirement_file)]),
                self.get_branch(),
                [update for updates in self.values() for update in updates if
                 self.should_update(update.requirement, update.requirement_file)]
            )

    @classmethod
    def get_branch(cls):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def get_body(cls, updates):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def get_empty_update_body(cls):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def get_title(cls):  # pragma: no cover
        raise NotImplementedError


class ScheduledUpdate(BundledUpdate):

    @classmethod
    def get_body(cls, updates):
        env = Environment(loader=FileSystemLoader(TEMPLATES_DIR))
        changelogs = [u.requirement for u in updates if u.requirement.changelog != {}]
        return env.get_template(
            "scheduled_update_body.md").render(
            {
                "updates": updates,
                "changelogs": changelogs,
                "api_key": settings.api_key
            }
        )

    def get_title(self):
        now = datetime.now()

        if "every day" in self.config.schedule:
            return "Scheduled daily dependency update on {}".format(now.strftime("%A"))
        elif "every week" in self.config.schedule:
            return "Scheduled weekly dependency update for week {}".format(now.strftime("%U"))
        elif "every two weeks" in self.config.schedule:
            return "Scheduled biweekly dependency update for week {}".format(now.strftime("%U"))
        elif "every month" in self.config.schedule:
            return "Scheduled monthly dependency update for {}".format(now.strftime("%B"))
        raise UnsupportedScheduleError("Unsupported schedule {}".format(self.config.schedule))

    def get_branch(self):
        return "scheduled-update-{dt}".format(
            dt=datetime.now().strftime("%Y-%m-%d")
        )


class InitialUpdate(BundledUpdate):

    @classmethod
    def get_body(cls, updates):
        env = Environment(loader=FileSystemLoader(TEMPLATES_DIR))
        changelogs = [u.requirement for u in updates if u.requirement.changelog != {}]
        return env.get_template(
            "initial_update_body.md"
        ).render(
            {
                "updates": updates,
                "changelogs": changelogs,
                "api_key": settings.api_key
            }
        )

    @classmethod
    def get_empty_update_body(cls):
        return "The initial setup worked, but all your packages are up to date. You can safely " \
               "close this issue."

    @classmethod
    def get_title(cls):
        return "Initial Update"

    @classmethod
    def get_branch(cls):
        return "initial-update"


RequirementUpdate = namedtuple(
    "RequirementUpdate",
    ["requirement_file", "requirement", "commit_message"]
)


class SequentialUpdate(Update):

    def get_updates(self):
        for key, updates in self.items():
            requirement, req_file = updates[0].requirement, updates[0].requirement_file
            if self.should_update(requirement, req_file):
                yield (
                    self.get_title(requirement),
                    self.get_body(requirement),
                    self.get_branch(requirement),
                    updates
                )

    @classmethod
    def get_branch(cls, requirement):
        if requirement.is_pinned:
            return "update-{}-{}-to-{}".format(
                requirement.key, requirement.version,
                requirement.latest_version_within_specs
            )
        return "pin-{}-{}".format(
            requirement.key,
            requirement.latest_version_within_specs
        )

    @classmethod
    def get_title(cls, requirement):
        if requirement.is_pinned:
            return "Update {} to {}".format(
                requirement.key,
                requirement.latest_version_within_specs
            )
        return "Pin {} to latest version {}".format(
            requirement.key,
            requirement.latest_version_within_specs
        )

    @classmethod
    def get_body(cls, requirement):
        env = Environment(loader=FileSystemLoader(TEMPLATES_DIR))
        return env.get_template("sequential_update_body.md").render({
            "requirement": requirement,
            "api_key": settings.api_key
        })



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import yaml
try:  # pragma: no cover
    basestring
except NameError:  # pragma: no cover
    basestring = str
import re


SCHEDULE_REGEX = re.compile(
    # has to begin with every
    r"^every "
    # followed by day/month
    "((day|month)$"
    # or week/two weeks
    "|(week|two weeks))"
    # with an optional weekday
    "( on (monday|tuesday|wednesday|thursday|friday|saturday|sunday))?",
    re.IGNORECASE
)


class Config(object):

    UPDATE_ALL = "all"
    UPDATE_INSECURE = "insecure"
    # the docs had a typo at some point that incorrectly reffered to 'security'
    # instead of 'insecure'.
    UPDATE_INSECURE_TYPO = "security"
    UPDATE_NONE = ["False", "false", False, None]

    def __init__(self):
        self.close_prs = True
        self.branch = "master"
        self.branch_prefix = "pyup-"
        self.pr_prefix = ""
        self.pin = True
        self.search = True
        self.requirements = []
        self.label_prs = False
        self.schedule = ""
        self.assignees = []
        self.gitlab = GitlabConfig()
        self.update = Config.UPDATE_ALL
        self.update_hashes = True

    def update_config(self, d):
        """
        Updates the config object.
        :param d: dict
        """
        for key, value in d.items():
            if hasattr(self, key):
                if key == "requirements":
                    items, value = value, []
                    for item in items:
                        if isinstance(item, basestring):
                            req = RequirementConfig(path=item)
                        elif isinstance(item, dict):
                            path, item = item.popitem()
                            req = RequirementConfig(
                                path=path,
                                pin=item.get("pin", None),
                                compile=item.get("compile", False),
                                update=item.get("update", Config.UPDATE_ALL)
                            )
                        value.append(req)
                        # add constraint requirement files to config
                        if req.compile:
                            for spec in req.compile.specs:
                                value.append(RequirementConfig(path=spec, pin=False))
                elif key == "assignees":
                    # assignees can be a string or a list. If it's a string, convert it to a list
                    # to make things consistent
                    if isinstance(value, basestring):
                        value = [value, ]
                elif key == 'gitlab':
                    value = GitlabConfig(**value)
                elif key == 'pr_prefix':
                    # make sure that pr prefixes don't contain a PIPE
                    if "|" in value:
                        continue
                # cast ints and floats to str
                if isinstance(value, (int, float)) and not isinstance(value, bool):
                    value = str(value)

                setattr(self, key, value)

    @staticmethod
    def generate_config_file(config):
        return "\n\n".join([
            "# autogenerated pyup.io config file \n# see https://pyup.io/docs/configuration/ "
            "for all available options",
            yaml.safe_dump(config, default_flow_style=False)
        ])

    def _get_requirement_attr(self, attr, path):
        """
        Gets the attribute for a given requirement file in path
        :param attr: string, attribute
        :param path: string, path
        :return: The attribute for the requirement, or the global default
        """
        for req_file in self.requirements:
            if path.strip("/") == req_file.path.strip("/"):
                return getattr(req_file, attr)
        return getattr(self, attr)

    def can_pin(self, path):
        """
        Checks if requirements in `path` can be pinned.
        :param path: string, path to requirement file
        :return: bool
        """
        return self._get_requirement_attr(attr="pin", path=path)

    def can_update_all(self, path):
        """
        Checks if requirements in `path` can be updated.
        :param path: string, path to requirement file
        :return: bool
        """
        return self._get_requirement_attr("update", path=path) == Config.UPDATE_ALL

    def can_update_insecure(self, path):
        """
        Checks if requirements in `path` can be updated if insecure.
        :param path: string, path to requirement file
        :return: bool
        """
        return self._get_requirement_attr("update", path=path) in (Config.UPDATE_ALL,
                                                                   Config.UPDATE_INSECURE,
                                                                   Config.UPDATE_INSECURE_TYPO)

    def is_valid_schedule(self):
        return SCHEDULE_REGEX.search(self.schedule) is not None

    def __repr__(self):
        return str(self.__dict__)


class RequirementConfig(object):

    def __init__(self, path, pin=None, compile=False, update=Config.UPDATE_ALL):
        self.path = path
        self.pin = pin
        self.compile = CompileConfig(specs=compile.get("specs", [])) if compile else False
        self.update = update

        # set pin default
        if self.pin is None:
            # don't pin pipfiles by default
            if self.path.endswith("Pipfile"):
                self.pin = False
            else:
                self.pin = True

    def __repr__(self):
        return str(self.__dict__)


class GitlabConfig(object):

    def __init__(self, should_remove_source_branch=False, merge_when_pipeline_succeeds=False):
        self.should_remove_source_branch = should_remove_source_branch
        self.merge_when_pipeline_succeeds = merge_when_pipeline_succeeds

    def __repr__(self):
        return str(self.__dict__)


class CompileConfig(object):

    def __init__(self, specs=list()):
        self.specs = specs

    def __repr__(self):
        return str(self.__dict__)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from unittest import TestCase
from pyup.pullrequest import PullRequest
from datetime import datetime, timedelta


def pullrequest_factory(title, state="open", url="http://foo.bar",
                        created_at=datetime.now(), number=1):
    return PullRequest(
        title=title,
        state=state,
        url=url,
        created_at=created_at,
        number=number
    )


class PullRequestTypeTest(TestCase):
    def test_update(self):
        pr = pullrequest_factory(title="Update this and that")
        self.assertEqual(pr.type, "update")
        self.assertTrue(pr.is_update)

    def test_security(self):
        pr = pullrequest_factory(title="Security is not provided here")
        self.assertEqual(pr.type, "security")
        self.assertTrue(pr.is_security)

    def test_pin(self):
        pr = pullrequest_factory(title="Pin this on thta")
        self.assertEqual(pr.type, "pin")
        self.assertTrue(pr.is_pin)

    def test_initial(self):
        pr = pullrequest_factory(title="Initial Update")
        self.assertEqual(pr.type, "initial")
        self.assertTrue(pr.is_initial)

    def test_unknown(self):
        pr = pullrequest_factory(title="Foo")
        self.assertEqual(pr.type, "unknown")
        #self.assertTrue(pr.is_scheduled)

    def test_compile(self):
        pr = pullrequest_factory(title="Compile foo.txt")
        self.assertEqual(pr.type, "compile")
        self.assertTrue(pr.is_compile)

    def test_scheduled(self):
        pr = pullrequest_factory(title="Scheduled foo in bar")
        self.assertEqual(pr.type, "scheduled")
        self.assertTrue(pr.is_scheduled)

    def test_is_valid(self):
        pr = pullrequest_factory(title="Scheduled foo in bar")
        self.assertTrue(pr.is_valid)

        pr = pullrequest_factory(title="Compile foo.txt")
        self.assertTrue(pr.is_valid)

        pr = pullrequest_factory(title="Initial Update")
        self.assertTrue(pr.is_valid)

        pr = pullrequest_factory(title="Pin this on thta")
        self.assertTrue(pr.is_valid)

        pr = pullrequest_factory(title="Update this and that")
        self.assertTrue(pr.is_valid)

class PullRequestEQTest(TestCase):
    def test_is_eq(self):
        pr1 = pullrequest_factory("yay", number=1)
        pr2 = pullrequest_factory("yay", number=2)
        self.assertNotEqual(pr1, pr2)
        pr1.number = pr2.number
        self.assertEqual(pr1, pr2)
        pr1.number = 3
        self.assertNotEqual(pr1, pr2)
        pr2.number = 3
        self.assertEqual(pr1, pr2)


class PullRequestStateTest(TestCase):

    def test_is_open(self):
        pr = pullrequest_factory(title="the foo", state="closed")
        self.assertEqual(pr.is_open, False)
        pr.state = "open"
        self.assertEqual(pr.is_open, True)


class PullRequestRequirementTestCase(TestCase):

    def test_initial_non(self):
        pr = pullrequest_factory(title="Initial PR")
        self.assertEqual(pr.get_requirement(), None)

    def test_pin(self):
        pr = pullrequest_factory(title="Pin django")
        self.assertEqual(pr.get_requirement(), "django")

    def test_update(self):
        pr = pullrequest_factory(title="Update django")
        self.assertEqual(pr.get_requirement(), "django")

    def test_some_bogus(self):
        pr = pullrequest_factory(title="Uhm?")
        self.assertEqual(pr.get_requirement(), None)

    def test_with_prefix(self):
        pr = pullrequest_factory(title="Some Prefix | Update django")
        self.assertEqual(pr.get_requirement("Some Prefix |"), 'django')

        flask = pullrequest_factory(title="Pin flask")
        flask_prefix = pullrequest_factory(title="Some Prefix | Pin flask")
        self.assertIsNotNone(flask.get_requirement())
        self.assertEqual(flask.get_requirement(), flask_prefix.get_requirement("Some Prefix |"))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
import logging
from gitlab import Gitlab
from gitlab.exceptions import GitlabGetError, GitlabCreateError
from ..errors import BranchExistsError, RepoDoesNotExistError
from base64 import b64encode

logger = logging.getLogger(__name__)


class BadTokenError(Exception):
    pass


class Provider(object):
    name = 'gitlab'

    class Committer(object):
        def __init__(self, login):
            self.login = login

    def __init__(self, bundle, intergration=False, url=None, ignore_ssl=False):
        self.bundle = bundle
        self.url = url
        self.ignore_ssl = ignore_ssl
        if intergration:
            raise NotImplementedError(
                'Gitlab provider does not support integration mode')

    @classmethod
    def is_same_user(cls, this, that):
        return this.login == that.login

    def _api(self, token):
        parts = token.split('@')
        if len(parts) == 1:
            host = self.url or 'https://gitlab.com'
            auth = parts[0]
        elif len(parts) == 2:
            auth, host = parts
        else:
            raise BadTokenError(
                'Got token "{}": format should be wither "apikey" for '
                'gitlab.com, or "apikey@https://yourgitlab.local"'.format(
                    token))
        return Gitlab(host, auth, ssl_verify=(not self.ignore_ssl))

    def get_user(self, token):
        gl = self._api(token)
        gl.auth()
        return gl.user

    def get_repo(self, token, name):
        try:
            return self._api(token).projects.get(name)
        except GitlabGetError as e:
            if e.response_code == 404:
                raise RepoDoesNotExistError()
            raise e

    def get_default_branch(self, repo):
        return repo.default_branch

    def get_pull_request_permissions(self, user, repo):
        # TODO: IDK how this works on gitlab
        return True

    def iter_git_tree(self, repo, branch):
        for item in repo.repository_tree(ref=branch, recursive=True, all=True):
            yield item['type'], item['path']

    def get_file(self, repo, path, branch):
        logger.info("Getting file at {} for branch {}".format(path, branch))
        # remove unnecessary leading slash to avoid gitlab errors. See #375
        path = path.lstrip('/')
        try:
            contentfile = repo.files.get(file_path=path, ref=branch)
        except GitlabGetError as e:
            if e.response_code == 404:
                logger.warning("Unable to get {path}".format(
                    path=path,
                ))
                return None, None
        else:
            return contentfile.decode().decode("utf-8"), contentfile

    def create_and_commit_file(self, repo, path, branch, content, commit_message, committer):

        # TODO: committer
        return repo.files.create({
            'file_path': path,
            'branch': branch,
            'content': content,
            'commit_message': commit_message
        })

    def get_requirement_file(self, repo, path, branch):
        content, file_obj = self.get_file(repo, path, branch)
        if content is not None and file_obj is not None:
            return self.bundle.get_requirement_file_class()(
                path=path,
                content=content,
            )
        return None

    def create_branch(self, repo, base_branch, new_branch):
        try:
            repo.branches.create({"branch": new_branch,
                                  "ref": base_branch})
        except GitlabCreateError as e:
            if e.error_message == 'Branch already exists':
                raise BranchExistsError(new_branch)

    def is_empty_branch(self, repo, base_branch, new_branch, prefix):
        """
        Compares the top commits of two branches.
        Please note: This function isn't checking if `base_branch` is a direct
        parent of `new_branch`, see
        http://stackoverflow.com/questions/3161204/find-the-parent-branch-of-a-git-branch
        :param repo: github.Repository
        :param base_branch: string name of the base branch
        :param new_branch: string name of the new branch
        :param prefix: string branch prefix, default 'pyup-'
        :return: bool -- True if empty
        """
        # extra safeguard to make sure we are handling a bot branch here
        assert new_branch.startswith(prefix)
        comp = repo.repository_compare(base_branch, new_branch)
        n = len(comp.commits)
        logger.info("Got a total of {} commits in {}".format(n, new_branch))
        return n == 0

    def delete_branch(self, repo, branch, prefix):
        """
        Deletes a branch.
        :param repo: github.Repository
        :param branch: string name of the branch to delete
        """
        # make sure that the name of the branch begins with pyup.
        assert branch.startswith(prefix)
        obj = repo.branches.get(branch)
        obj.delete()

    def create_commit(self, path, branch, commit_message, content, sha, repo, committer):
        # TODO: committer

        f = repo.files.get(file_path=path, ref=branch)
        # Gitlab supports a plaintext encoding, which is when the encoding
        # value is unset.  Python-Gitlab seems to set it to b64, so we can't
        # unset it unfortunately
        f.content = b64encode(content.encode()).decode()
        f.encoding = 'base64'
        # TODO: committer
        f.save(branch=branch, commit_message=commit_message)

    def get_pull_request_committer(self, repo, pull_request):
        return [
            self.Committer(participant['username'])
            for participant in repo.mergerequests.get(pull_request.number).participants()
        ]

    def close_pull_request(self, bot_repo, user_repo, pull_request, comment, prefix):
        mr = user_repo.mergerequests.get(pull_request.number)
        mr.state_event = 'close'
        mr.save()
        mr.notes.create({'body': comment})

        source_branch = mr.changes()['source_branch']
        logger.info("Deleting source branch {}".format(source_branch))
        self.delete_branch(user_repo, source_branch, prefix)

    def _merge_merge_request(self, mr, config):
        mr.merge(remove_source_branch=config.gitlab.should_remove_source_branch,
                 merge_when_pipeline_succeeds=True)

    def create_pull_request(self, repo, title, body, base_branch, new_branch, pr_label, assignees, config):
        # TODO: Check permissions
        try:
            if len(body) >= 65536:
                logger.warning("PR body exceeds maximum length of 65536 chars, reducing")
                body = body[:65536 - 1]

            mr = repo.mergerequests.create({
                'source_branch': new_branch,
                'target_branch': base_branch,
                'title': title,
                'description': body,
                'pr_label': pr_label,
                'remove_source_branch': config.gitlab.should_remove_source_branch
            })

            if config.gitlab.merge_when_pipeline_succeeds:
                self._merge_merge_request(mr, config)

            return self.bundle.get_pull_request_class()(
                state=mr.state,
                title=mr.title,
                url=mr.web_url,
                created_at=mr.created_at,
                number=mr.iid,
                issue=False
            )
        except GitlabCreateError as e:
            if e.response_code == 409:
                logger.warning(
                    "PR {title} from {base_branch}->{new_branch} is already "
                    "exists and is open, resorting to a comment "
                    "instead".format(
                        title=title, new_branch=new_branch,
                        base_branch=base_branch))

                comment = '# {title}\n{body}'.format(title=title, body=body)
                # the exception doesn't say *which* MR is open, so we have to
                # find it :(
                for mr in repo.mergerequests.list(state='opened', all=True):
                    if (mr.source_branch == new_branch and mr.target_branch == base_branch):
                        mr.notes.create({'body': comment})
                        return self.bundle.get_pull_request_class()(
                            state=mr.state,
                            title=mr.title,
                            url=mr.web_url,
                            created_at=mr.created_at,
                            number=mr.iid,
                            issue=False
                        )

    def create_issue(self, repo, title, body):
        return repo.issues.create({
            'title': title,
            'description': body
        })

    def iter_issues(self, repo, creator):
        # TODO: handle creator
        for issue in repo.mergerequests.list(state='opened', all=True):
            yield self.bundle.get_pull_request_class()(
                state=issue.state,
                title=issue.title,
                url=issue.web_url,
                created_at=issue.created_at,
                number=issue.iid,
                issue=True,
            )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from unittest import TestCase
from pyup.requirements import Requirement
from mock import patch, PropertyMock, Mock
from pyup.requirements import RequirementFile, RequirementsBundle
from .test_package import package_factory
import requests_mock
import os


class RequirementUpdateContent(TestCase):

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_tox_ini(self, _):
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "[testenv:bandit]\n" \
                      "commands =\n" \
                      "\tbandit --ini setup.cfg -ii -l --recursive project_directory\n" \
                      "deps =\n" \
                      "\tbandit==1.4.0\n" \
                      "\n" \
                      "[testenv:manifest]\n" \
                      "commands =\n" \
                      "\tcheck-manifest --verbose\n"
            req_file = RequirementFile("tox.ini", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("bandit==1.4.0", 1),
                req
            )
            updated = req.update_content(content)
            new_content = "[testenv:bandit]\n" \
                      "commands =\n" \
                      "\tbandit --ini setup.cfg -ii -l --recursive project_directory\n" \
                      "deps =\n" \
                      "\tbandit==2.9.5\n" \
                      "\n" \
                      "[testenv:manifest]\n" \
                      "commands =\n" \
                      "\tcheck-manifest --verbose\n"
            self.assertEqual(updated, new_content)

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_conda_file(self, _):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "name: my_env\n" \
                      "dependencies:\n" \
                      "  - gevent=1.2.1\n" \
                      "  - pip:\n" \
                      "    - beautifulsoup4==1.2.3\n"
            req_file = RequirementFile("req.yml", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("beautifulsoup4==1.2.3", 1),
                req
            )
            updated = req.update_content(content)
            new_content = "name: my_env\n" \
                      "dependencies:\n" \
                      "  - gevent=1.2.1\n" \
                      "  - pip:\n" \
                      "    - beautifulsoup4==2.9.5\n"
            self.assertEqual(updated, new_content)

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_multispace(self, _):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "                   pass"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("pass", 1),
                req
            )
            updated = req.update_content(content)
            self.assertEqual(updated, "pass==2.9.5")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_compatible(self, _):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "Jinja2~=2.9.4         # via flask"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("Jinja2~=2.9.4         # via flask", 1),
                req
            )

            self.assertEqual(req.update_content(content), "Jinja2==2.9.5         # via flask")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_compatible_matching_latest(self, _):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "Jinja2~=2.9.5         # via flask"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("Jinja2~=2.9.5         # via flask", 1),
                req
            )

            self.assertEqual(req.update_content(content), "Jinja2==2.9.5         # via flask")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_update_contains_correct_sep_char(self, _):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="2.9.5"):
            content = "Jinja2==2.9.4         # via flask"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("Jinja2==2.9.4         # via flask", 1),
                req
            )

            self.assertEqual(req.update_content(content), "Jinja2==2.9.5         # via flask")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9 \\\n" \
                      "        --hash=sha256:abcde"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9", 1),
                req
            )

            self.assertEqual(req.update_content(content), "alembic==1.4.2 \\\n"
                                                          "    --hash=sha256:123 \\\n"
                                                          "    --hash=sha256:456")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_sorted(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "abc"}, {"hash": "456"},
                                        {"hash": "789"}, {"hash": "123"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9 \\\n" \
                      "        --hash=sha256:abcde"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9", 1),
                req
            )

            self.assertEqual(req.update_content(content), "alembic==1.4.2 \\\n"
                                                          "    --hash=sha256:123 \\\n"
                                                          "    --hash=sha256:456 \\\n"
                                                          "    --hash=sha256:789 \\\n"
                                                          "    --hash=sha256:abc")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_and_comment(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9 \\\n" \
                      "        --hash=sha256:abcde # yay"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "alembic==1.4.2 \\\n"
                                                          "    --hash=sha256:123 \\\n"
                                                          "    --hash=sha256:456 # yay")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_and_comment_and_env_markers(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9; sys_platform != 'win32' \\\n" \
                      "        --hash=sha256:abcde # yay"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9; sys_platform != 'win32'", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "alembic==1.4.2; sys_platform != 'win32' \\\n"
                                                          "    --hash=sha256:123 \\\n"
                                                          "    --hash=sha256:456 # yay")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_and_comment_all_in_one_line(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9 --hash=sha256:abcde # yay"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "alembic==1.4.2 \\\n"
                                          "    --hash=sha256:123 \\\n"
                                          "    --hash=sha256:456 # yay")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_taskcluster_215(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "taskcluster==0.3.4 --hash sha256:d4fe5e2a44fe27e195b92830ece0a6eb9eb7ad9dc556a0cb16f6f2a6429f1b65"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("taskcluster==0.3.4", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "taskcluster==1.4.2 \\\n"
                                          "    --hash=sha256:123 \\\n"
                                          "    --hash=sha256:456")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_and_comment_and_env_markers_all_in_one_line(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9; sys_platform != 'win32' --hash=sha256:abcde # yay"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9; sys_platform != 'win32'", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "alembic==1.4.2; sys_platform != 'win32' \\\n"
                                          "    --hash=sha256:123 \\\n"
                                          "    --hash=sha256:456 # yay")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    @patch("pyup.requirements.Requirement.get_hashes")
    def test_update_with_hashes_in_one_line(self, get_hashes_mock, _):
        get_hashes_mock.return_value = [{"hash": "123"}, {"hash": "456"}]
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "alembic==0.8.9 --hash=sha256:abcde"
            req_file = RequirementFile("req.txt", content)
            req = list(req_file.requirements)[0]
            self.assertEqual(
                Requirement.parse("alembic==0.8.9", 1),
                req
            )

            new_content = req.update_content(content)
            self.assertEqual(new_content, "alembic==1.4.2 \\\n"
                                          "    --hash=sha256:123 \\\n"
                                          "    --hash=sha256:456")

    def test_update_with_environment_markers(self):
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):

            content = "uvloop==0.6.5; sys_platform != 'win32'"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "uvloop==1.4.2; sys_platform != 'win32'")

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_update_with_environment_markers_and_comment(self, _):
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock,
                   return_value="1.4.2"):

            content = "uvloop==0.6.5; sys_platform != 'win32' # and here's some comment"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "uvloop==1.4.2; sys_platform != 'win32' # and here's some comment")

    def test_update_content_with_extras(self):
        with patch('pyup.requirements.Requirement.latest_version_within_specs', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "requests[security]==1.4.1"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "requests[security]==1.4.2")

    def test_update_content_tabbed(self):
        with patch('pyup.requirements.Requirement.latest_version_within_specs', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "bla==1.4.1\t\t# some foo"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "bla==1.4.2\t\t# some foo")

            content = "bla==1.4.1\t\t# pyup: <1.4.2"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "bla==1.4.2\t\t# pyup: <1.4.2")

    def test_something_else(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="0.13.1"):
            content = "some-package==0.12.2+tmf"
            req = Requirement.parse(content, 0)
            self.assertEqual(req.update_content(content), "some-package==0.13.1")

    def test_line_endings(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.2.3"):
            with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                       return_value=package_factory("Foo", [])):
                content = """\r\n\r\nWerkzeug\r\ndjango-template-repl\nbpython\nsome-fooo    \n"""
                r = RequirementFile("foo.txt", content)
                self.assertEqual(r.requirements[0].name, "Werkzeug")
                self.assertEqual(r.requirements[1].name, "django-template-repl")
                self.assertEqual(r.requirements[2].name, "bpython")
                self.assertEqual(r.requirements[3].name, "some-fooo")
                self.assertTrue("Werkzeug==1.2.3\r\n" in r.requirements[0].update_content(content))
                self.assertTrue(
                    "django-template-repl==1.2.3\n" in r.requirements[1].update_content(content))
                self.assertTrue(
                    "bpython==1.2.3" in r.requirements[2].update_content(content))
                self.assertTrue(
                    "some-fooo==1.2.3    \n" in r.requirements[3].update_content(content))

    def test_update_content_simple_pinned(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "Django==1.4.1"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "Django==1.4.2")

        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "django==1.4.1"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "django==1.4.2")

    def test_latest_version_within_specs_called(self):

        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock, return_value="1.4.2") as mocked:
            content = "django==1.4.1"
            req = Requirement.parse(content, 0)
            self.assertEqual(req.update_content(content), "django==1.4.2")
            mocked.assert_called_with()

    def test_update_content_simple_unpinned(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "django"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "django==1.4.2")

        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "Django"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "Django==1.4.2")

    def test_update_content_simple_unpinned_with_comment(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "django # newest django release"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "django==1.4.2 # newest django release")

        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "Django #django"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content), "Django==1.4.2 #django")

        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.4.2"):
            content = "Django #django #yay this has really cool comments ######"
            req = Requirement.parse(content, 0)

            self.assertEqual(req.update_content(content),
                             "Django==1.4.2 #django #yay this has really cool comments ######")

    def test_update_content_with_package_in_comments(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="2.58.1.44"):
            content = 'raven==5.8.1\n' \
                      '{%- endif %}\n\n' \
                      '{% if cookiecutter.use_newrelic == "y" -%}\n' \
                      '# Newrelic agent for performance monitoring\n' \
                      '# -----------------------------------------\n' \
                      'newrelic\n' \
                      '{%- endif %}\n\n'
            req = Requirement.parse("newrelic", 0)
            updated_content = 'raven==5.8.1\n' \
                              '{%- endif %}\n\n' \
                              '{% if cookiecutter.use_newrelic == "y" -%}\n' \
                              '# Newrelic agent for performance monitoring\n' \
                              '# -----------------------------------------\n' \
                              'newrelic==2.58.1.44\n' \
                              '{%- endif %}\n\n'
            self.assertEqual(req.update_content(content), updated_content)

    def test_update_content_with_dubious_package_name(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="2.58.1.44"):
            content = 'raven\n' \
                      'ravenclient'
            req = Requirement.parse("raven", 0)
            updated_content = 'raven==2.58.1.44\n' \
                              'ravenclient'
            self.assertEqual(req.update_content(content), updated_content)

    def test_update_content_ranged(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.5.6"):
            content = 'raven>=0.2\n' \
                      'ravenclient'
            req = Requirement.parse("raven>=0.2", 0)
            updated_content = 'raven==1.5.6\n' \
                              'ravenclient'
            self.assertEqual(req.update_content(content), updated_content)

    def test_update_content_unfinished_line(self):
        with patch('pyup.requirements.Requirement.latest_version', new_callable=PropertyMock,
                   return_value="1.5.6"):
            content = 'raven==0.2\n'
            req = Requirement.parse("raven==0.2", 0)
            updated_content = 'raven==1.5.6\n'
            self.assertEqual(req.update_content(content), updated_content)


class RequirementTestCase(TestCase):

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_is_outdated(self, _):
        with patch('pyup.requirements.Requirement.latest_version_within_specs',
                   new_callable=PropertyMock, return_value=None):
            r = Requirement.parse("Django", 0)
            self.assertEqual(r.is_outdated, False)

    def test_equals(self):
        self.assertEqual(
            Requirement.parse("Django==1.5", 0),
            Requirement.parse("Django==1.5", 0)
        )

    def test_not_equals(self):
        self.assertNotEqual(
            Requirement.parse("Django==1.5", 0),
            Requirement.parse("Django==1.6", 0)
        )

    def test_filter(self):
        from packaging.specifiers import SpecifierSet
        r = Requirement.parse("Django==1.7.6", 0)
        self.assertEqual(r.filter, False)

        r = Requirement.parse("Django==1.7.6 # pyup: < 1.7.8", 0)
        self.assertEqual(r.filter, SpecifierSet("<1.7.8"))

        req = Requirement.parse("some-package==1.9.3 # rq.filter: <1.10 #some comment here", 0)
        self.assertEqual(req.filter, SpecifierSet("<1.10"))

        r = Requirement.parse("django==1.7.1  # pyup: <1.7.6", 0)

        r = Requirement.parse("Django==1.7.6 # pyup: < 1.7.8, > 1.7.2", 0)
        self.assertEqual(
            sorted([s._spec for s in r.filter._specs], key=lambda r: r[1]),
            sorted([("<", "1.7.8"), (">", "1.7.2")], key=lambda r: r[1])
        )

        r = Requirement.parse("Django==1.7.6 # pyup:< 1.7.8", 0)
        self.assertEqual(r.filter, SpecifierSet("<1.7.8"))

    def test_until_filter(self):
        from packaging.specifiers import SpecifierSet
        from datetime import datetime, timedelta
        tommorow = (datetime.now() + timedelta(days=1)).strftime("%Y-%m-%d")
        yesterday = (datetime.now() - timedelta(days=1)).strftime("%Y-%m-%d")

        r = Requirement.parse("Django==1.7.6 # pyup: < 1.7.8 until {}".format(yesterday), 0)
        self.assertEqual(r.filter, False)

        r = Requirement.parse("Django==1.7.6 # pyup: < 1.7.8 until {}".format(tommorow), 0)
        self.assertEqual(r.filter, SpecifierSet("<1.7.8"))

        r = Requirement.parse("Django==1.7.6 # pyup: < 1.7.8 until 348-34-1234", 0)
        self.assertEqual(r.filter, SpecifierSet("<1.7.8"))


    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_update_filter(self, _):
        latest_fn = 'pyup.requirements.Requirement.latest_version'

        with patch(latest_fn, new_callable=PropertyMock, return_value='1.8.7'):
            r = Requirement.parse("Django==1.7.6", 0)
            self.assertTrue(r.needs_update)
            self.assertTrue(r.can_update_semver)

            r = Requirement.parse("Django==1.7.6 # pyup: update major", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

            r = Requirement.parse("Django==0.2.6 # pyup: update major", 0)
            self.assertTrue(r.can_update_semver)
            self.assertTrue(r.needs_update)

        with patch(latest_fn, new_callable=PropertyMock, return_value='1.8.7'):
            r = Requirement.parse("Django==1.7.6", 0)
            self.assertTrue(r.needs_update)
            self.assertTrue(r.can_update_semver)

            r = Requirement.parse("Django==1.7.6 # pyup: update minor", 0)
            self.assertTrue(r.can_update_semver)
            self.assertTrue(r.needs_update)

            r = Requirement.parse("Django==1.8.6 # pyup: update minor", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

            r = Requirement.parse("Django==1.9.6 # pyup: update minor", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

        with patch(latest_fn, new_callable=PropertyMock, return_value='1.2.4'):
            r = Requirement.parse("Django==1.2.3", 0)
            self.assertTrue(r.needs_update)
            self.assertTrue(r.can_update_semver)

            r = Requirement.parse("Django==1.2.3 # pyup: update minor", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

            r = Requirement.parse("Django==1.2.2 # pyup: update minor", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

            r = Requirement.parse("Django==1.2 # pyup: update minor", 0)
            self.assertFalse(r.can_update_semver)
            self.assertFalse(r.needs_update)

    def test_convert_semver(self):
        self.assertEqual({"major": 1, "minor": 2, "patch": 3}, Requirement.convert_semver("1.2.3"))
        self.assertEqual({"major": 1, "minor": 2, "patch": 0}, Requirement.convert_semver("1.2"))
        self.assertEqual({"major": 1, "minor": 0, "patch": 0}, Requirement.convert_semver("1"))

    def test_tabbed(self):
        req = Requirement.parse("Django==1.5\t\t#some-comment", 0)
        self.assertEqual(req.is_pinned, True)
        self.assertEqual(req.version, "1.5")

    def test_is_pinned(self):
        req = Requirement.parse("Django", 0)
        self.assertEqual(req.is_pinned, False)

        req = Requirement.parse("Django==1.4,>1.5", 0)
        self.assertEqual(req.is_pinned, False)

        req = Requirement.parse("Django===1.4", 0)
        self.assertEqual(req.is_pinned, False)

        req = Requirement.parse("Django<=1.4,>=1.33", 0)
        self.assertEqual(req.is_pinned, False)

        req = Requirement.parse("Django==1.4", 0)
        self.assertEqual(req.is_pinned, True)

    def test_is_loose(self):
        req = Requirement.parse("Django", 0)
        self.assertEqual(req.is_loose, True)

        req = Requirement.parse("Django==1.4,>1.5", 0)
        self.assertEqual(req.is_loose, False)

        req = Requirement.parse("Django===1.4", 0)
        self.assertEqual(req.is_loose, False)

        req = Requirement.parse("Django<=1.4,>=1.33", 0)
        self.assertEqual(req.is_loose, False)

        req = Requirement.parse("Django==1.4", 0)
        self.assertEqual(req.is_loose, False)

    def test_is_open_ranged(self):
        req = Requirement.parse("Django>=1.5", 0)
        self.assertEqual(req.is_open_ranged, True)

        req = Requirement.parse("Django<=1.4,>1.5", 0)
        self.assertEqual(req.is_open_ranged, False)

        req = Requirement.parse("Django===1.4", 0)
        self.assertEqual(req.is_open_ranged, False)

        req = Requirement.parse("Django<=1.4,>=1.33", 0)
        self.assertEqual(req.is_open_ranged, False)

        req = Requirement.parse("Django==1.4", 0)
        self.assertEqual(req.is_open_ranged, False)

        req = Requirement.parse("Django>=1.4,<=1.7", 0)
        self.assertEqual(req.is_open_ranged, False)

    def test_package_filter_present(self):
        req = Requirement.parse("Django", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("Django #rq.filter:", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("Django #rq.filter: >=1.4,<1.5", 0)
        self.assertEqual(
            sorted([s._spec for s in req.filter._specs], key=lambda i: i[0]),
            sorted([('>=', '1.4'), ('<', '1.5')], key=lambda i: i[0])
        )

        req = Requirement.parse("Django #rq.filter:!=1.2", 0)
        self.assertEqual([s._spec for s in req.filter._specs], [('!=', '1.2')])

        req = Requirement.parse("Django #rq.filter:foo", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("bliss #rq.filter:", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("Django", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("Django #pyup:", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("Django #pyup: >=1.4,<1.5", 0)
        self.assertEqual(
            sorted([s._spec for s in req.filter._specs], key=lambda i: i[0]),
            sorted([('>=', '1.4'), ('<', '1.5')], key=lambda i: i[0])
        )


        req = Requirement.parse("Django #pyup:!=1.2", 0)
        self.assertEqual([s._spec for s in req.filter._specs], [('!=', '1.2')])

        req = Requirement.parse("Django #pyup:foo", 0)
        self.assertEqual(req.filter, False)

        req = Requirement.parse("bliss #pyup:", 0)
        self.assertEqual(req.filter, False)


    def test_get_latest_version_within_specs(self):
        from packaging.specifiers import SpecifierSet
        latest = Requirement.get_latest_version_within_specs(
            SpecifierSet("==1.2,!=1.2"),
            ["1.2", "1.3", "1.4", "1.5"]
        )

        self.assertEqual(latest, None)

        latest = Requirement.get_latest_version_within_specs(
            SpecifierSet("==1.2.1"),
            ["1.2.0", "1.2.1", "1.2.2", "1.3"]
        )

        self.assertEqual(latest, "1.2.1")

    def test_latest_version_within_specs(self):
        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory("bliss",
                                                versions=["1.9rc1", "1.9", "1.8.1", "1.8", "1.7",
                                                          "1.6"])):
            req = Requirement.parse("bliss #rq.filter:", 0)
            self.assertEqual(req.latest_version_within_specs, "1.9")

            req = Requirement.parse("bliss==1.8rc1 #rq.filter:", 0)
            self.assertEqual(req.prereleases, True)
            self.assertEqual(req.latest_version_within_specs, "1.9rc1")

            req = Requirement.parse("bliss #rq.filter: >=1.7,<1.9", 0)
            self.assertEqual(req.latest_version_within_specs, "1.8.1")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory("gevent",
                                                versions=['1.1rc1', '1.1b6', '1.1b5', '1.1b4',
                                                          '1.1b3', '1.1b2', '1.1b1', '1.1a2',
                                                          '1.1a1', '1.0.2', '1.0.1', ])):
            req = Requirement.parse("gevent==1.1b6", 0)
            self.assertEqual(req.latest_version_within_specs, "1.1rc1")
            self.assertEqual(req.latest_version, "1.1rc1")

    def test_version_unpinned(self):
        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django", versions=["1.9", "1.8"])):
            r = Requirement.parse("Django", 0)
            self.assertEqual(r.version, "1.9")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django", versions=["1.9rc1", "1.9", "1.8"])):
            r = Requirement.parse("Django", 0)
            self.assertEqual(r.version, "1.9")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django",
                                                versions=["1.9.1", "1.8", "1.9rc1"])):
            r = Requirement.parse("django", 0)
            self.assertEqual(r.version, "1.9.1")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(
                       name="django",
                       versions=["1.4.3", "1.5", "1.4.2", "1.4.1", ])):
            r = Requirement.parse("Django  # rq.filter: >=1.4,<1.5", 0)
            self.assertEqual(r.version, "1.4.3")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(
                       name="django",
                       versions=["1.4.3", "1.5", "1.4.2", "1.4.1", ])):
            r = Requirement.parse("Django  # pyup: >=1.4,<1.5", 0)
            self.assertEqual(r.version, "1.4.3")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django", versions=["1.8.1", "1.8"])):
            r = Requirement.parse("Django  # rq.filter: !=1.8.1", 0)
            self.assertEqual(r.version, "1.8")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django", versions=["1.8.1", "1.8"])):
            r = Requirement.parse("Django  # pyup: !=1.8.1", 0)
            self.assertEqual(r.version, "1.8")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django",
                                                versions=["1.9rc1", "1.9.1", "1.8", ])):
            r = Requirement.parse("django  # rq.filter: bogus", 0)
            self.assertEqual(r.version, "1.9.1")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django",
                                                versions=["1.9rc1", "1.9.1", "1.8", ])):
            r = Requirement.parse("django  # pyup: bogus", 0)
            self.assertEqual(r.version, "1.9.1")

    def test_version_pinned(self):
        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django", versions=["1.8", "1.9"])):
            r = Requirement.parse("Django==1.9", 0)
            self.assertEqual(r.version, "1.9")

        with patch('pyup.requirements.Requirement.package', new_callable=PropertyMock,
                   return_value=package_factory(name="django==1.9rc1",
                                                versions=["1.8", "1.9rc1", "1.9rc2"])):
            r = Requirement.parse("Django==1.9.2.rc14 # rq.filter != 1.44", 0)
            self.assertEqual(r.version, "1.9.2.rc14")

    def test_version_compatible(self):
        req = Requirement.parse("Django~=1.5", 0)
        self.assertEqual(
            sorted([s._spec for s in req.specs._specs], key=lambda i: i[0]),
            sorted([('>=', '1.5'), ('<', '2.0')], key=lambda i: i[0])
        )

        req = Requirement.parse("Django~=1.5.2", 0)
        self.assertEqual(
            sorted([s._spec for s in req.specs._specs], key=lambda i: i[0]),
            sorted([('>=', '1.5.2'), ('<', '1.6.0')], key=lambda i: i[0])
        )

    def test_prereleases(self):
        r = Requirement.parse("Django==1.9rc1", 0)
        self.assertEqual(r.prereleases, True)

        r = Requirement.parse("Django==1.9-b1", 0)
        self.assertEqual(r.prereleases, True)

        r = Requirement.parse("Django==1.9-alpha1", 0)
        self.assertEqual(r.prereleases, True)

        r = Requirement.parse("Django", 0)
        self.assertEqual(r.prereleases, False)

        r = Requirement.parse("Django>=1.5,<=1.6", 0)
        self.assertEqual(r.prereleases, False)

        r = Requirement.parse("Django!=1.9", 0)
        self.assertEqual(r.prereleases, False)

    def test_name(self):
        r = Requirement.parse("Django==1.9rc1", 0)
        self.assertEqual(r.name, "Django")

        r = Requirement.parse("django==1.9-b1", 0)
        self.assertEqual(r.name, "django")

    @requests_mock.mock()
    def test_package_found(self, requests):
        with open(os.path.dirname(os.path.realpath(__file__)) + "/data/django.json") as f:
            requests.get("https://pypi.org/pypi/Django/json", text=f.read())
        r = Requirement.parse("Django==1.9rc1", 0)
        self.assertEqual(r._fetched_package, False)
        self.assertEqual(r._package, None)

        # this triggers the fetch
        self.assertNotEqual(r.package, None)
        self.assertEqual(r._fetched_package, True)
        self.assertNotEqual(r._package, None)

    @requests_mock.mock()
    def test_package_not_found(self, requests):
        requests.get("https://pypi.org/pypi/Fango/json", text="404", status_code=404)
        r = Requirement.parse("Fango", 0)
        self.assertEqual(r._fetched_package, False)
        self.assertEqual(r._package, None)

        # this triggers the fetch
        self.assertEqual(r.package, None)
        self.assertEqual(r._fetched_package, True)
        self.assertEqual(r._package, None)

    @patch("pyup.requirements.safety")
    @patch("pyup.requirements.settings")
    def test_is_insecure(self, settings, safety):
        """with self.assertRaises(NotImplementedError):
            r = Requirement.parse("Django", 0)
            r.is_insecure"""
        # test safety is not called if the api key not set
        settings.api_key = None
        r = Requirement.parse("Django", 0)
        self.assertFalse(r.is_insecure)
        safety.check.assert_not_called()

        # test safety is called if the api key is set
        settings.api_key = "foo"
        r = Requirement.parse("Django", 0)
        self.assertFalse(r.is_insecure)

        safety.check.return_value = []
        safety.check.assert_called_with(
            cached=True,
            db_mirror='',
            ignore_ids=(),
            key=settings.api_key,
            packages=(r,)
        )

    @requests_mock.mock()
    @patch("pyup.requirements.settings")
    def test_changelogs(self, requests, settings):
        # test that the api is not called if the api key is not set
        settings.api_key = None
        r = Requirement.parse("pyupio", 0)
        self.assertEqual(r.changelog, {})

        # test the api is called if the api key is set
        settings.api_key = "foo"
        r = Requirement.parse("pyupio", 0)
        from pyup.package import Package
        r._package = Package("pyupio", versions=['0.2'])
        r._fetched_package = True
        with open(os.path.dirname(os.path.realpath(__file__)) + "/data/pyup-changelog.json") as f:
            requests.get("https://pyup.io/api/v1/changelogs/pyupio/", text=f.read())
            log = r.changelog
            self.assertEqual(len(log), 6)

    @requests_mock.mock()
    def test_needs_update(self, requests):
        with open(os.path.dirname(os.path.realpath(__file__)) + "/data/django.json") as f:
            requests.get("https://pypi.org/pypi/Django/json", text=f.read())

            # is pinned and on latest
            r = Requirement.parse("Django==1.9rc1", 0)
            self.assertEqual(r.needs_update, False)

            # is ranged and open
            r = Requirement.parse("Django>=1.8", 0)
            self.assertEqual(r.needs_update, False)

            # is range and not open
            r = Requirement.parse("Django>=1.8.2,<1.9.0", 0)
            self.assertEqual(r.needs_update, False)

            # is pinned but old
            r = Requirement.parse("Django==1.7", 0)
            self.assertEqual(r.needs_update, True)

            # is not pinned
            r = Requirement.parse("Django", 0)
            self.assertEqual(r.needs_update, True)

            # is compatible and on latest
            r = Requirement.parse("Django~=1.8.2", 0)
            self.assertEqual(r.needs_update, False)

            # is compatible but old
            r = Requirement.parse("Django~=1.7.0", 0)
            self.assertEqual(r.needs_update, True)

    def test_str(self):
        r = Requirement.parse("Django==1.9rc1", 0)
        self.assertEqual(r.__str__(), "Requirement.parse(Django==1.9rc1, 0)")


class RequirementsFileTestCase(TestCase):

    def test_parse_empty_line(self):
        r = RequirementFile("foo.txt", "\n\n\n\n\n")
        self.assertEqual(r.requirements, [])

    @patch("pyup.requirements.Requirement.package", return_value="pkg")
    def test_parse_hashes(self, _):
        with open(os.path.dirname(os.path.realpath(__file__)) + "/data/hashed_reqs.txt") as f:
            content = f.read()
            r = RequirementFile("r.txt", content)
            self.assertEqual(
                [
                    Requirement.parse("alembic==0.8.9", 1),
                    Requirement.parse("amqp==2.1.1", 3),
                    Requirement.parse("argon2-cffi==16.3.0", 6),
                    Requirement.parse("Babel==2.3.4", 8),
                ],
                list(r.requirements)
            )

    def test_parse_index_server(self):
        line = "--index-url https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "-i https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url https://some.foo"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url https://some.foo # some lousy comment"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "-i\t\t https://some.foo \t\t    # some lousy comment"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--index-url"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            None
        )

        line = "--index-url=https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "-i=https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url=https://some.foo/"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url=https://some.foo"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "--extra-index-url=https://some.foo # some lousy comment"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

        line = "-i\t\t =https://some.foo \t\t    # some lousy comment"
        self.assertEqual(
            RequirementFile.parse_index_server(line),
            "https://some.foo/"
        )

    @patch("pyup.requirements.Requirement.package")
    def test_parse_package_with_index_server(self, package):
        content = """-i https://some.foo/\ndjango"""
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r.requirements[0].index_server, "https://some.foo/")

    def test_parse_empty_line(self):
        content = """
        """
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r.requirements, [])
        self.assertEqual(r._other_files, [])

    def test_parse_comment_line(self):
        content = """
# the comment is here
        """
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r.requirements, [])
        self.assertEqual(r._other_files, [])

    def test_unsupported_line_start(self):
        content = """
-f foo
--find-links bla
-i bla
--index-url bla
--extra-index-url bla
--no-index bla
--allow-external
--allow-unverified
-Z
--always-unzip
        """
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r.requirements, [])
        self.assertEqual(r._other_files, [])

    @patch("pyup.requirements.Requirement.package")
    def test_parse_requirement(self, package):
        package.return_value = True
        content = """
-e common/lib/calc
South==1.0.1
pycrypto>=2.6
git+https://github.com/pmitros/pyfs.git@96e1922348bfe6d99201b9512a9ed946c87b7e0b
distribute>=0.6.28, <0.7
# bogus comment
-e .
pdfminer==20140328
-r production/requirements.txt
--requirement test.txt
        """
        r = RequirementFile("r.txt", content=content)

        self.assertEqual(
            r.other_files, [
                "production/requirements.txt",
                "test.txt"
            ]
        )

        self.assertEqual(
            r.requirements, [
                Requirement.parse("South==1.0.1", 3),
                Requirement.parse("pycrypto>=2.6", 4),
                Requirement.parse("distribute>=0.6.28, <0.7", 6),
                Requirement.parse("pdfminer==20140328", 3),
            ]
        )

    @patch("pyup.requirements.Requirement.package")
    def test_ignore_file(self, package):
        package.return_value = True
        content = """# pyup: ignore file
foo
bar
baz"""
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r.requirements, [])
        self.assertEqual(r._other_files, [])
        self.assertEqual(r.is_valid, False)

    @patch("pyup.requirements.Requirement.package")
    def test_ignore_requirement(self, package):
        package.return_value = True

        content = """foo
bar # pyup: ignore
baz"""
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(len(r.requirements), 2)
        self.assertEqual(
            r.requirements, [
                Requirement.parse("foo", 0),
                Requirement.parse("baz", 2)
            ]
        )

        content = """foo
        bar # pyup:ignore
        baz"""
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(len(r.requirements), 2)
        self.assertEqual(
            r.requirements, [
                Requirement.parse("foo", 0),
                Requirement.parse("baz", 2)
            ]
        )

    def test_resolve_file(self):
        resolved = RequirementFile.resolve_file(
            "base/requirements.txt",
            "-r requirements/production.txt"
        )
        self.assertEqual(resolved, "base/requirements/production.txt")

        resolved = RequirementFile.resolve_file(
            "base/requirements.txt",
            "-r requirements/production.txt # prod file"
        )
        self.assertEqual(resolved, "base/requirements/production.txt")

        resolved = RequirementFile.resolve_file(
            "requirements.txt",
            "-r production.txt # prod file"
        )
        self.assertEqual(resolved, "production.txt")

    def test_is_invalid(self):
        content = ''
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r._is_valid, None)

        # this triggers the _parse
        self.assertEqual(r.is_valid, False)
        self.assertEqual(r._is_valid, False)

    def test_is_valid_other_file(self):
        content = '-r other_file.txt'
        r = RequirementFile("r.txt", content=content)
        self.assertEqual(r._is_valid, None)

        # this triggers the _parse
        self.assertEqual(r.is_valid, True)
        self.assertEqual(r._is_valid, True)

    def test_is_valid_requirement(self):
        with patch('pyup.requirements.Requirement.package', return_value=True):
            content = 'some_package'
            r = RequirementFile("r.txt", content=content)
            self.assertEqual(r._is_valid, None)

            # this triggers the _parse
            self.assertEqual(r.is_valid, True)
            self.assertEqual(r._is_valid, True)

    def test_str(self):
        r = RequirementFile("r.txt", "content", "asdfe")
        self.assertEqual(
            r.__str__(),
            "RequirementFile(path='r.txt', sha='asdfe', content='content')"
        )

        content = "more content than 30 characters here"
        r = RequirementFile("r.txt", content, "asdfe")
        self.assertEqual(
            r.__str__(),
            "RequirementFile(path='r.txt', sha='asdfe', "
            "content='more content than 30 character[truncated]')"
        )


class RequirementsBundleTestCase(TestCase):
    def test_has_file(self):
        reqs = RequirementsBundle()
        self.assertEqual(reqs.has_file_in_path("foo.txt"), False)
        self.assertEqual(reqs.has_file_in_path(""), False)
        reqs.append(RequirementFile(path="foo.txt", content=''))
        self.assertEqual(reqs.has_file_in_path("foo.txt"), True)

    def test_add(self):
        reqs = RequirementsBundle()
        self.assertEqual(reqs, [])
        reqs.append(RequirementFile(path="foo.txt", content=''))
        self.assertEqual(reqs[0].path, "foo.txt")

    def test_get_initial_update_class(self):
        req = RequirementsBundle()
        klass = req.get_update_class(
            initial=True,
            scheduled=False,
            config=None
        )
        self.assertEqual(klass, req.get_initial_update_class())

    def test_get_scheduled_update_class(self):
        req = RequirementsBundle()
        config = Mock()
        config.is_valid_schedule.return_value = True
        klass = req.get_update_class(
            initial=False,
            scheduled=True,
            config=config
        )
        self.assertEqual(klass, req.get_scheduled_update_class())

    def test_get_sequential_update_class(self):
        req = RequirementsBundle()
        klass = req.get_update_class(
            initial=False,
            scheduled=False,
            config=None
        )
        self.assertEqual(klass, req.get_sequential_update_class())

    def test_get_updates(self):
        with patch('pyup.requirements.Requirement.package', return_value=Mock()):
            reqs = RequirementsBundle()
            reqs.append(RequirementFile(path="r.txt", content='Bla'))
            updates = [u for u in reqs.get_updates(True, False, Mock())]
            self.assertEqual(len(updates), 1)
            #self.assertEqual(updates[0].__class__, reqs.get_initial_update_class().__class__)

            reqs = RequirementsBundle()
            reqs.append(RequirementFile(path="r.txt", content='Bla'))
            updates = [u for u in reqs.get_updates(False, False, Mock())]
            self.assertEqual(len(updates), 1)
            #self.assertEqual(updates[0].__class__, reqs.get_sequential_update_class().__class__)

    def test_requirements(self):
        with patch('pyup.requirements.Requirement.package', return_value=Mock()):
            reqs = RequirementsBundle()
            reqs.append(RequirementFile(path="r.txt", content='Bla\nFoo'))

            self.assertEqual([
                Requirement.parse("Bla", 1),
                Requirement.parse("Foo", 2)
            ],
                [r for r in reqs.requirements]
            )




# -*- coding: utf-8 -*-
from pyup import __version__, settings
from pyup.bot import Bot
from pyup.requirements import RequirementFile, RequirementsBundle
from pyup.providers.github import Provider as GithubProvider
from pyup.providers.gitlab import Provider as GitlabProvider

import click
from tqdm import tqdm
import logging


@click.command()
@click.version_option(__version__, '-v', '--version')
@click.option('--repo', prompt='repository', help='')
@click.option('--user-token', prompt='user token', help='')
@click.option('--bot-token', help='', default=None)
@click.option("--key", default="",
              help="API Key for pyup.io's vulnerability database. Can be set as SAFETY_API_KEY "
                   "environment variable. Default: empty")
@click.option('--provider', help='API to use; either github or gitlab', default="github")
@click.option('--provider_url', help='Optional custom URL to your provider', default=None)
@click.option('--branch', help='Set the branch the bot should use', default='master')
@click.option('--initial', help='Set this to bundle all PRs into a large one',
              default=False, is_flag=True)
@click.option('--ignore_ssl', help='Set this to ignore SSL Certificate',
              default=False, is_flag=True)
@click.option('--log', help='Set the log level', default="ERROR")
def main(repo, user_token, bot_token, key, provider, provider_url, branch, initial, ignore_ssl, log):
    logging.basicConfig(level=getattr(logging, log.upper(), None))

    settings.configure(key=key)

    if provider == 'github':
        ProviderClass = GithubProvider
    elif provider == 'gitlab':
        ProviderClass = GitlabProvider
    else:
        raise NotImplementedError

    bot = CLIBot(
        repo=repo,
        user_token=user_token,
        bot_token=bot_token,
        provider=ProviderClass,
        provider_url=provider_url,
        ignore_ssl=ignore_ssl,
    )

    bot.update(branch=branch, initial=initial)


if __name__ == '__main__':
    main()


class CLIBot(Bot):

    def __init__(self, repo, user_token, bot_token=None,
                 provider=GithubProvider, bundle=RequirementsBundle,
                 provider_url=None, ignore_ssl=False):
        bundle = CLIBundle
        super(CLIBot, self).__init__(repo, user_token, bot_token, provider,
                                     bundle, provider_url=provider_url,
                                     ignore_ssl=ignore_ssl)

    def iter_updates(self, initial, scheduled):

        ls = list(super(CLIBot, self).iter_updates(initial, scheduled))

        if not initial:
            ls = tqdm(ls, desc="Updating ...")
        for title, body, update_branch, updates in ls:
            if not initial:
                ls.set_description(title)
            yield title, body, update_branch, updates

    def iter_changes(self, initial, updates):
        # we don't display the progress bar if this is a sequential update, just return the list
        if initial:
            updates = tqdm(updates, desc="Updating ...")
        for update in updates:
            if initial:
                updates.set_description(update.commit_message)
            yield update


class CLIBundle(RequirementsBundle):

    def get_requirement_file_class(self):
        return CLIRequirementFile


class CLIRequirementFile(RequirementFile):
    def iter_lines(self, lineno=0):
        bar = tqdm(self.content.splitlines()[lineno:], desc="Processing {}".format(self.path))
        for item in bar:
            yield item



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from unittest import TestCase
from pyup.bot import Bot
from .test_pullrequest import pullrequest_factory
from pyup.updates import RequirementUpdate, InitialUpdate
from pyup.requirements import RequirementFile
from pyup.errors import NoPermissionError, ConfigError
from pyup.config import RequirementConfig
from mock import Mock, patch


def bot_factory(repo="foo/foo", user_token="foo", bot_token=None,
                bot_class=Bot, ignore_ssl=False, prs=list()):
    bot = bot_class(
        repo=repo,
        user_token=user_token,
        bot_token=bot_token,
        ignore_ssl=ignore_ssl,
    )
    bot._fetched_prs = True
    bot.req_bundle.pull_requests = prs
    bot.provider = Mock()
    bot.config.update_config({
        "close_prs": True,
        "pin": True,
        "branch": "base_branch",
        "search": True
    })
    return bot


class BotUserRepoTest(TestCase):
    def test_user_repo(self):
        bot = bot_factory()
        bot.provider.get_repo.return_value = "THE REPO"

        self.assertEqual(bot.user_repo, "THE REPO")


class BotUserTest(TestCase):
    def test_user(self):
        bot = bot_factory()
        bot.provider.get_user.return_value = "THE USER"

        self.assertEqual(bot.user, "THE USER")


class BotBotTest(TestCase):
    def test_bot_without_token(self):
        bot = bot_factory()
        bot.provider.get_user.return_value = "THE BOT"
        self.assertEqual(bot.bot, "THE BOT")

    def test_bot_with_token(self):
        bot = bot_factory(bot_token="the foo")
        bot.provider.get_user.return_value = "THE BOT"
        self.assertEqual(bot.bot, "THE BOT")


class BotBotRepoTest(TestCase):
    def test_bot_repo(self):
        bot = bot_factory()
        bot.provider.get_repo.return_value = "THE BOT REPO"
        self.assertEqual(bot.bot_repo, "THE BOT REPO")


class BotPullRequestsTest(TestCase):

    def test_iter_issues_called(self):
        bot = bot_factory()
        bot._fetched_prs = False
        bot.provider.iter_issues = Mock(return_value=[])
        bot.pull_requests
        self.assertEqual(bot.provider.iter_issues.call_count, 1)


class BotRepoConfigTest(TestCase):

    def test_fetches_file_success(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = "foo: bar", None
        self.assertEqual(bot.get_repo_config(bot.user_repo), {"foo": "bar"})

    def test_yaml_error(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = "foo: bar: baz: fii:", None
        with self.assertRaises(ConfigError):
            self.assertEqual(bot.get_repo_config(bot.user_repo), None)

    def test_fetches_file_error(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = None, None
        self.assertEqual(bot.get_repo_config(bot.user_repo), None)


class BotConfigureTest(TestCase):

    def test_kwargs(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = None, None
        bot.configure(branch="bogus-branch", pin="bogus-pin", close_prs="bogus-close")
        self.assertEqual(bot.config.branch, "bogus-branch")
        self.assertEqual(bot.config.pin, "bogus-pin")
        self.assertEqual(bot.config.close_prs, "bogus-close")

    def test_file(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = "close_prs: bogus-close\nbranch: bogus-branch", None
        bot.configure()
        self.assertEqual(bot.config.branch, "bogus-branch")
        self.assertEqual(bot.config.close_prs, "bogus-close")

    def test_numeric_branch(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = "branch: 2.0\n", None
        bot.configure()
        self.assertEqual(bot.config.branch, "2.0")

    def test_write_config(self):
        bot = bot_factory()
        bot.provider.get_file.return_value = None, None
        bot.configure(write_config={'branch': 'bogus-branch'})
        self.assertEqual(bot.config.branch, "bogus-branch")


class BotUpdateTest(TestCase):
    def test_branch_is_none(self):
        bot = bot_factory()
        bot.provider.get_default_branch.return_value = "the foo"
        bot.provider.get_file.return_value = None, None
        bot.get_all_requirements = Mock()
        bot.apply_updates = Mock()
        bot.update()
        self.assertEqual(bot.config.branch, "the foo")

    def test_branch_is_set(self):
        bot = bot_factory()
        bot.get_all_requirements = Mock()
        bot.apply_updates = Mock()
        bot.provider.get_file.return_value = None, None
        bot.update(branch="the branch")
        self.assertEqual(bot.config.branch, "the branch")


class BotApplyUpdateTest(TestCase):
    def test_apply_update_pull_request_exists(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("The PR")

        bot = bot_factory(prs=[the_pull])
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [
            ("The PR", "", "", [update])]
        bot.apply_updates(initial=True, scheduled=False)

        self.assertEqual(the_requirement.pull_request, the_pull)

    def test_updates_empty(self):
        bot = bot_factory()
        bot.create_issue = Mock()
        bot.req_bundle.get_updates = Mock(side_effect=IndexError)
        bot.apply_updates(initial=True, scheduled=False)
        bot.create_issue.assert_called_once_with(
            title=InitialUpdate.get_title(),
            body=InitialUpdate.get_empty_update_body()
        )

    def test_updates_empty_with_prefix(self):
        bot = bot_factory()
        bot.config.pr_prefix = "Some Prefix"
        bot.create_issue = Mock()
        bot.req_bundle.get_updates = Mock(side_effect=IndexError)
        bot.apply_updates(initial=True, scheduled=False)
        bot.create_issue.assert_called_once_with(
            title="Some Prefix " + InitialUpdate.get_title(),
            body=InitialUpdate.get_empty_update_body()
        )

    def test_updates_empty_with_write_config(self):
        bot = bot_factory()
        bot.write_config = {'foo': 'bar'}
        bot.create_issue = Mock()
        bot.req_bundle.get_updates = Mock(side_effect=IndexError)
        bot.pull_config = Mock()
        bot.apply_updates(initial=True, scheduled=False)
        bot.create_issue.assert_called_once_with(
            title=InitialUpdate.get_title(),
            body=InitialUpdate.get_empty_update_body()
        )
        bot.pull_config.assert_called_once_with(
            {'foo': 'bar'}
        )

    def test_apply_update_pull_request_new(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("The PR")

        bot = bot_factory(prs=[the_pull])
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update])]
        bot.commit_and_pull = Mock()
        bot.commit_and_pull.return_value = the_pull
        bot.apply_updates(initial=True, scheduled=False)

        self.assertEqual(the_requirement.pull_request, the_pull)

    def test_apply_update_with_prefix_pull_request_new(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("The PR")

        bot = bot_factory(prs=[the_pull])
        bot.config.pr_prefix = "Some Prefix"
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update])]
        bot.commit_and_pull = Mock()
        bot.commit_and_pull.return_value = the_pull
        bot.apply_updates(initial=False, scheduled=False)
        self.assertEqual(the_requirement.pull_request, the_pull)
        bot.commit_and_pull.assert_called_once_with(
            body='',
            initial=False,
            new_branch=u'pyup-',
            title=u'Some Prefix The PR',
            updates=[update]
        )

    def test_apply_update_with_write_config(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("The PR")

        bot = bot_factory(prs=[the_pull])
        bot.write_config = {'foo': 'bar'}
        bot.pull_config = Mock()
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update])]
        bot.commit_and_pull = Mock()
        bot.commit_and_pull.return_value = the_pull
        bot.apply_updates(initial=True, scheduled=False)

        self.assertEqual(the_requirement.pull_request, the_pull)
        bot.pull_config.assert_called_once_with(
            {'foo': 'bar'}
        )

    def test_close_stall_prs_called(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("The PR")
        bot = bot_factory(prs=[])
        bot.close_stale_prs = Mock()
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update])]
        bot.commit_and_pull = Mock()
        bot.commit_and_pull.return_value = the_pull
        bot.apply_updates(initial=False, scheduled=False)

        self.assertEqual(the_requirement.pull_request, the_pull)
        bot.close_stale_prs.assert_called_once_with(update=update, pull_request=the_pull,
                                                    scheduled=False)

    def test_close_stall_prs_called_only_once_on_scheduled_run(self):
        the_requirement = Mock()
        the_pull = pullrequest_factory("Scheduled")
        bot = bot_factory(prs=[])
        bot.close_stale_prs = Mock()
        bot.req_bundle.get_updates = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update, update])]
        bot.commit_and_pull = Mock()
        bot.commit_and_pull.return_value = the_pull
        bot.apply_updates(initial=False, scheduled=True)

        self.assertEqual(the_requirement.pull_request, the_pull)
        bot.close_stale_prs.assert_called_once_with(update=update, pull_request=the_pull,
                                                    scheduled=True)

    def test_apply_update_initial_empty(self):
        bot = bot_factory()
        bot.req_bundle.get_updates = Mock()
        bot.req_bundle.get_updates.return_value = [("", "", "", [])]
        bot.provider.create_issue.return_value = None
        bot.apply_updates(initial=True, scheduled=False)

        create_issue_args_list = bot.provider.create_issue.call_args_list
        self.assertEqual(len(create_issue_args_list), 1)
        self.assertEqual(
            create_issue_args_list[0][1]["body"],
            InitialUpdate.get_empty_update_body()
        )
        self.assertEqual(
            create_issue_args_list[0][1]["title"],
            InitialUpdate.get_title()
        )

    def test_apply_update_initial_pr_still_open(self):
        initial_pr = pullrequest_factory(
            title=InitialUpdate.get_title(),
            state="open",
        )
        bot = bot_factory(prs=[initial_pr])
        the_requirement = Mock()
        update = RequirementUpdate(
            requirement_file="foo", requirement=the_requirement, commit_message="foo"
        )
        bot.req_bundle.get_updates = Mock()
        bot.req_bundle.get_updates.return_value = [("The PR", "", "", [update])]

        bot.apply_updates(initial=True, scheduled=False)

        self.assertEqual(bot.provider.create_pull_request.called, False)


class BotCommitAndPullTest(TestCase):
    def test_multiple_updates_in_file(self):
        bot = bot_factory()
        bot.provider.create_branch = Mock()
        bot.provider.create_commit.side_effect = [
            "sha1", "sha2", "sha3"
        ]
        bot.create_pull_request = Mock()
        requirement = Mock()
        requirement.update_content.return_value = "new content"
        updates = [
            RequirementUpdate(
                requirement_file=RequirementFile(
                    path="foo.txt",
                    content='',
                    sha='abcd'
                ),
                requirement=requirement,
                commit_message="foo"
            ),
            RequirementUpdate(
                requirement_file=RequirementFile(
                    path="foo.txt",
                    content='',
                    sha='abcd'
                ),
                requirement=requirement,
                commit_message="foo"
            ),
            RequirementUpdate(
                requirement_file=RequirementFile(
                    path="baz.txt",
                    content='',
                    sha='xyz'
                ),
                requirement=requirement,
                commit_message="foo"
            )
        ]

        bot.commit_and_pull(True, "new branch", "repo", "", updates)

        self.assertEqual(bot.provider.create_commit.called, True)
        self.assertEqual(bot.provider.create_commit.call_count, 2)
        create_commit_calls = bot.provider.create_commit.call_args_list
        # we're looking for the sha here. Make sure that the sha got updated with the new content
        self.assertEqual(create_commit_calls[0][1]["sha"], "abcd")
        self.assertEqual(create_commit_calls[1][1]["sha"], "xyz")

    def test_create_branch_fails(self):
        bot = bot_factory()
        bot.create_branch = Mock(return_value=False)
        self.assertEqual(bot.commit_and_pull(None, None, None, None, None), None)


class CreateBranchTest(TestCase):

    def test_success(self):
        bot = bot_factory()
        self.assertEqual(bot.create_branch("new-branch", delete_empty=False), True)
        bot.provider.create_branch.assert_called_once_with(
            base_branch="base_branch", new_branch="new-branch", repo=bot.user_repo)

    def test_error_dont_delete(self):
        from pyup.errors import BranchExistsError
        bot = bot_factory()
        bot.provider.create_branch.side_effect = BranchExistsError
        self.assertEqual(bot.create_branch("new-branch", delete_empty=False), False)
        bot.provider.is_empty_branch.assert_not_called()
        bot.provider.delete_branch.assert_not_called()

    def test_error_delete(self):
        from pyup.errors import BranchExistsError
        bot = bot_factory()
        bot.provider.create_branch.side_effect = BranchExistsError
        bot.provider.is_empty_branch.return_value = True
        bot.create_branch("new-branch", delete_empty=True)

        self.assertEqual(bot.provider.is_empty_branch.call_count, 1)
        self.assertEqual(bot.provider.delete_branch.call_count, 1)
        self.assertEqual(len(bot.provider.create_branch.mock_calls), 2)

    def test_branch_not_empty(self):
        from pyup.errors import BranchExistsError
        bot = bot_factory()
        bot.provider.create_branch.side_effect = BranchExistsError
        bot.provider.is_empty_branch.return_value = False
        bot.create_branch("new-branch", delete_empty=True)

        self.assertEqual(bot.provider.is_empty_branch.call_count, 1)
        bot.provider.delete_branch.assert_not_called()
        self.assertEqual(len(bot.provider.create_branch.mock_calls), 1)



class BotGetAllRequirementsTest(TestCase):
    def test_non_matching_file_not_added(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "foo.py"),  # not added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, False)

    def test_requirement_not_in_path(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "this/that/bla/dev.pip"),  # not added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, False)

    def test_file_not_ending_with_txt_or_pip(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "requirements/dev"),  # not added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, False)

    def test_matching_file_deep(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "requirements/dev.txt"),  # added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, True)

    def test_matching_file(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "requirements.txt"),  # added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, True)

    def test_matching_file_pip(self):
        bot = bot_factory()
        bot.provider.iter_git_tree.return_value = ("blob", "requirements.pip"),  # added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, True)

    def test_no_search(self):
        bot = bot_factory()
        bot.config.search = False
        bot.provider.iter_git_tree.return_value = ("blob", "requirements.pip"),  # added
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, False)

    def test_requirement_in_config(self):
        bot = bot_factory()
        bot.config.search = False
        bot.config.requirements = [
            RequirementConfig(path="foo.txt")
        ]
        bot.add_requirement_file = Mock()
        bot.get_all_requirements()
        self.assertEqual(bot.add_requirement_file.called, True)
        bot.add_requirement_file.assert_called_once_with("foo.txt", sha=None)


class BotAddRequirementFileTest(TestCase):
    def test_file_is_in_path(self):
        bot = bot_factory()
        bot.req_bundle.has_file_in_path = Mock()
        bot.req_bundle.append = Mock()
        bot.req_bundle.has_file_in_path.return_value = True

        bot.add_requirement_file("path",)

        self.assertEqual(bot.provider.get_requirement_file.called, False)
        self.assertEqual(bot.req_bundle.append.called, False)

    def test_file_not_found(self):
        bot = bot_factory()
        bot.req_bundle.has_file_in_path = Mock()
        bot.req_bundle.append = Mock()
        bot.provider.get_requirement_file.return_value = None
        bot.req_bundle.has_file_in_path.return_value = False

        bot.add_requirement_file("path",)

        self.assertEqual(bot.provider.get_requirement_file.called, True)
        self.assertEqual(bot.req_bundle.append.called, False)

    def test_file_found_single(self):
        bot = bot_factory()
        bot.req_bundle.has_file_in_path = Mock()
        bot.req_bundle.append = Mock()
        req_file = RequirementFile("path", "")
        bot.provider.get_requirement_file.return_value = req_file

        bot.req_bundle.has_file_in_path.return_value = False

        bot.add_requirement_file("path",)

        self.assertEqual(bot.provider.get_requirement_file.called, True)
        self.assertEqual(bot.req_bundle.append.called, True)

    def test_file_found_with_reference(self):
        bot = bot_factory()
        bot.req_bundle.has_file_in_path = Mock()
        bot.req_bundle.append = Mock()
        req_file = RequirementFile("path", "-r foo.txt")
        bot.provider.get_requirement_file.side_effect = [req_file, None]

        bot.req_bundle.has_file_in_path.return_value = False

        bot.add_requirement_file("path")

        self.assertEqual(bot.provider.get_requirement_file.called, True)
        self.assertEqual(bot.req_bundle.append.called, True)


class BotCanPullTest(TestCase):

    def test_valid_schedule_but_unscheduled_run(self):
        bot = bot_factory(bot_token=None)
        bot.config.is_valid_schedule = Mock()
        bot.config.is_valid_schedule.return_value = True
        self.assertFalse(bot.can_pull(False, False))

    def test_valid_schedule_and_scheduled_run(self):
        bot = bot_factory(bot_token=None)
        bot.config.is_valid_schedule = Mock()
        bot.config.is_valid_schedule.return_value = True
        self.assertTrue(bot.can_pull(False, True))

    def test_no_schedule(self):
        bot = bot_factory(bot_token=None)
        bot.config.is_valid_schedule = Mock()
        bot.config.is_valid_schedule.return_value = False
        self.assertTrue(bot.can_pull(False, False))
        self.assertTrue(bot.can_pull(False, True))

    def test_initial(self):
        bot = bot_factory(bot_token=None)
        bot.config.is_valid_schedule = Mock()
        bot.config.is_valid_schedule.return_value = False
        self.assertTrue(bot.can_pull(True, False))
        self.assertTrue(bot.can_pull(True, True))


class BotCreatePullRequestTest(TestCase):

    def test_plain(self):
        bot = bot_factory(bot_token=None)
        bot._bot_repo = "BOT REPO"
        bot._user_repo = "USER REPO"
        bot.create_pull_request("title", "body", "new_branch")
        self.assertEqual(bot.provider.create_pull_request.called, True)
        self.assertEqual(bot.provider.create_pull_request.call_args_list[0][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "USER REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config
        })

    def test_bot_no_errors(self):
        bot = bot_factory(bot_token="foo")
        bot._bot_repo = "BOT REPO"
        bot._user_repo = "USER REPO"
        bot.create_pull_request("title", "body", "new_branch")
        self.assertEqual(bot.provider.create_pull_request.called, True)
        self.assertEqual(bot.provider.create_pull_request.call_args_list[0][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "BOT REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config
        })
        self.assertEqual(bot.provider.get_pull_request_permissions.called, False)

    def test_bot_permission_error_resolved(self):
        bot = bot_factory(bot_token="foo")
        bot.provider.create_pull_request.side_effect = [NoPermissionError, "the foo"]
        bot._bot_repo = "BOT REPO"
        bot._user_repo = "USER REPO"
        bot.create_pull_request("title", "body", "new_branch")
        self.assertEqual(bot.provider.create_pull_request.called, True)
        self.assertEqual(bot.provider.create_pull_request.call_args_list[0][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "BOT REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config
        })
        self.assertEqual(bot.provider.create_pull_request.call_args_list[1][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "BOT REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config

        })

    def test_bot_permission_error_not_resolved(self):
        bot = bot_factory(bot_token="foo")
        bot.provider.create_pull_request.side_effect = [NoPermissionError, NoPermissionError]
        bot._bot_repo = "BOT REPO"
        bot._user_repo = "USER REPO"
        with self.assertRaises(NoPermissionError):
            bot.create_pull_request("title", "body", "new_branch")
        self.assertEqual(bot.provider.create_pull_request.called, True)
        self.assertEqual(bot.provider.create_pull_request.call_args_list[0][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "BOT REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config
        })
        self.assertEqual(bot.provider.create_pull_request.call_args_list[1][1], {
            "base_branch": "base_branch",
            "new_branch": "new_branch",
            "repo": "BOT REPO",
            "body": "body",
            "title": "title",
            "pr_label": False,
            "assignees": [],
            "config": bot.config
        })


class CloseStalePRsTestCase(TestCase):

    def setUp(self):

        self.pr = Mock()
        self.pr.title = "First PR"
        self.pr.canonical_title.return_value = "First PR"
        self.pr.number = 100
        self.pr.type = "update"
        self.pr.is_update = True
        self.pr.is_initial = False

        self.update = Mock()
        self.update.requirement.key = "some-req"

        self.other_pr = Mock()
        self.other_pr.type = "update"
        self.other_pr.is_open = True
        self.other_pr.title = "Second PR"
        self.other_pr.canonical_title.return_value = "Second PR"
        self.other_pr.get_requirement.return_value = "some-req"
        self.other_pr.is_update = True
        self.other_pr.is_initial = False

    def test_scheduled_closing_scheduled(self):
        self.pr.is_scheduled = True
        self.other_pr.is_scheduled = True
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, True)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo,
                                                                        self.other_pr)

    def test_scheduled_closing_update(self):
        self.pr.is_scheduled = True
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, True)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo,
                                                                        self.other_pr)

    def test_no_bot_token(self):
        bot = bot_factory()
        self.pr.type = Mock()
        bot.close_stale_prs(self.update, self.pr, False)

        self.assertEqual(self.pr.type.call_count, 0)

    def test_no_pull_requests(self):
        bot = bot_factory(bot_token="foo")

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_not_called()

    def test_close_success(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo, self.other_pr)
        bot.provider.close_pull_request.assert_called_once_with(
            bot_repo=bot.bot_repo,
            user_repo=bot.user_repo,
            pull_request=self.other_pr,
            comment="Closing this in favor of #100",
            prefix="pyup-"
        )

    def test_close_integration(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        bot.integration = True
        bot.provider.integration = True
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo, self.other_pr)
        bot.provider.close_pull_request.assert_called_once_with(
            bot_repo=bot.bot_repo,
            user_repo=bot.user_repo,
            pull_request=self.other_pr,
            comment="Closing this in favor of #100",
            prefix="pyup-"
        )

    def test_close_success_with_prefix(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        bot.config.pr_prefix = "Some Prefix"
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo, self.other_pr)
        bot.provider.close_pull_request.assert_called_once_with(
            bot_repo=bot.bot_repo,
            user_repo=bot.user_repo,
            pull_request=self.other_pr,
            comment="Closing this in favor of #100",
            prefix="pyup-"
        )

    def test_wrong_pr_type(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        self.other_pr.is_update = False
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_not_called()
        bot.provider.close_pull_request.assert_not_called()

    def test_pr_closed(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        self.other_pr.is_open = False
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_not_called()
        bot.provider.close_pull_request.assert_not_called()

    def test_same_title(self):

        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        self.other_pr.title = "First PR"
        self.other_pr.canonical_title.return_value = "First PR"
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_not_called()
        bot.provider.close_pull_request.assert_not_called()


    def test_requirement_doesnt_match(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        self.other_pr.get_requirement.return_value = "other-req"
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_not_called()
        bot.provider.close_pull_request.assert_not_called()

    def test_more_than_one_committer(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        commiter, commiter1 = Mock(), Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter, commiter1]

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo, self.other_pr)
        bot.provider.close_pull_request.assert_not_called()

    def test_committer_is_not_bot_user(self):
        bot = bot_factory(bot_token="foo", prs=[self.other_pr])
        commiter = Mock()
        bot.provider.get_pull_request_committer.return_value = [commiter]
        bot.provider.is_same_user.return_value = False

        bot.close_stale_prs(self.update, self.pr, False)

        bot.provider.get_pull_request_committer.assert_called_once_with(bot.user_repo, self.other_pr)
        bot.provider.close_pull_request.assert_not_called()


class ConflictingUpdateTest(TestCase):

    def test_no_conflict(self):
        bot = bot_factory()
        update1 = Mock()
        update1.requirement.key = "pkg"
        update1.requirement.latest_version_within_specs = "1.0"

        update2 = Mock()
        update2.requirement.key = "other-pkg"
        update1.requirement.latest_version_within_specs = "1.0"

        bot.iter_updates = Mock(return_value=[
            [None, None, None, [update1]],
            [None, None, None, [update2]]
        ])

        self.assertFalse(
            bot.has_conflicting_update(update1)
        )

    def test_has_conflict(self):
        bot = bot_factory()
        update1 = Mock()
        update1.requirement.key = "pkg"
        update1.requirement.latest_version_within_specs = "1.0"

        update2 = Mock()
        update2.requirement.key = "pkg"
        update1.requirement.latest_version_within_specs = "1.4"

        bot.iter_updates = Mock(return_value=[
            [None, None, None, [update1]],
            [None, None, None, [update2]]
        ])

        self.assertTrue(
            bot.has_conflicting_update(update1)
        )

    def test_fool_loop(self):
        bot = bot_factory()
        update1 = Mock()
        update1.requirement.key = "google-api-python-client"
        update1.requirement.latest_version_within_specs = "1.5.3"
        update1.commit_message = "Update google-api-python-client from 1.5.1 to 1.5.3"

        update2 = Mock()
        update2.requirement.key = "google-api-python-client"
        update2.requirement.latest_version_within_specs = "1.5.3"
        update2.commit_message = "Pin google-api-python-client to latest version 1.5.3"

        bot.iter_updates = Mock(return_value=[
            [None, None, None, [update1]],
            [None, None, None, [update2]]
        ])

        self.assertTrue(
            bot.has_conflicting_update(update1)
        )


class IgnoreSslTest(TestCase):
    def test_ignore_ssl_default_false(self):
        bot = Bot(repo='foo/foo', user_token='foo')
        self.assertFalse(bot.provider.ignore_ssl)

    def test_ignore_ssl_true(self):
        bot = Bot(repo='foo/foo', user_token='foo', ignore_ssl=True)
        self.assertTrue(bot.provider.ignore_ssl)
